
# This file was generated by 'versioneer.py' (0.19) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2021-03-19T15:36:52+0900",
 "dirty": true,
 "error": null,
 "full-revisionid": "a501b6fc3f3fb56f217250f1f2100715af24699a",
 "version": "0.9.0+1.ga501b6f.dirty"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
